--[[
Name: LibSpreadsheet-1.0
Revision: $Rev: 21 $
Author(s): Humbedooh
Description: Spreadsheet stuff
Dependencies: LibStub AceGUI-3.0
License: GPL v3 or later.
]]
-----------------------------------------------------------------------

local SPREADSHEET = "LibSpreadsheet-1.0"
local SPREADSHEET_MINOR = tonumber(("$Rev: 021 $"):match("(%d+)"))
if not LibStub then error(SPREADSHEET .. " requires LibStub.") end
local lib = LibStub:NewLibrary(SPREADSHEET, SPREADSHEET_MINOR)
if not lib then return end


--[[
LibSpreadsheet-1.0 DEFAULTS
]]--
lib.disabled = lib.disabled or nil;
lib.objects = lib.objects or {};
lib.str = UIParent:CreateFontString();
lib.cols = {'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z', '?'};
lib.sortAscending = 1;
lib.sortDescending = 2;
lib.sortAscendingWithCase = 3;
lib.sortDescendingWithCase = 4;
lib.callbackEvents = { 'OnSelectionChange', 'OnUserInputProcessed', 'OnUserInputRequested', 'OnUserSortRequested', 'OnUserSortProcessed' }

lib.fonts = {
    StringFont = "LibSpreadsheet_StringFont",
    NumberFont = "LibSpreadsheet_NumberFont",
    HeaderFont = "LibSpreadsheet_HeaderFont",
    ChineseHeaderFont = "LibSpreadsheet_ChineseHeaderFont",
};




--[[
LibSpreadsheet-1.0 BOOK API
]]--

function lib:Book(title)
	local obj = {};
	if ( type(title) ~= "string" ) then error(("%s: Error in creating book title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
    
    -- Book default values
	obj.sheets =            {};
    obj.parent =            self;
    obj.selectedSheet =     nil;
	obj.type =              'book';
	obj.title =             title;
    obj.strict =            false;
    obj.hasBookUI =         false;
    obj.hasSelector =       false;
    obj.editbox =           nil;
    
    -- Book functions
	obj.Add =               lib.AddSheet;
	obj.Delete =            lib.DeleteSheet;
	obj.Select =            lib.Select;
	obj.StrictMode =        lib.StrictMode;
    obj.ShowSelector =      lib.ShowSelector;
    obj.ShowScrollbars =    lib.ShowScrollbars;
    obj.Leaf =              lib.Leaf;
    obj.Render =            lib.Render;
    obj.Hide =              lib.Hide;
    
    obj.locale =            GetLocale();
    if ( obj.locale == "zhCN" or obj.locale == "zhTW" ) then
        lib.fonts.StringFont = "LibSpreadsheet_ChineseStringFont";
        lib.fonts.HeaderFont = "LibSpreadsheet_ChineseHeaderFont";
    end
	return obj;
end


--[[
LibSpreadsheet-1.0 SHEET API
]]--

function lib:AddSheet(title)
	local obj = self;
	if ( type(title) ~= "string" ) then error(("%s: Error in setting sheet title; Expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then error(("%s: Sheet \"%s\" already exists in book \"%s\"!"):format(SPREADSHEET, title, obj.title)); end
	end
	local sheet = {
        parent =                    obj,
		type =                      'sheet',
		title =                     title,
        callbacks =                 {},
		columns =                   {},
        sortDirection =             obj.parent.sortAscending,
        lastColumn =                0,
		rows =                      {},
		rowHeight =                 0,
		lastRow =                   0,
		lastColumn =                0,
        currentRow =                nil,
        currentColumn =             nil,
		frame =                     nil,
        isInteractive =             false,
        cellSpacing =               2,
        
		AddColumn =                 lib.AddColumn,
		SetColumn =                 lib.SetColumn,
		DeleteColumn =              lib.DeleteColumn,	
        ExtractLocation =           lib.ExtractLocation,
		AddRow =                    lib.AddRow,
        DeleteRow =                 lib.DeleteRow,
		GoTo =                      lib.GoTo,
		Insert =                    lib.Insert,
		Delete =                    lib.Delete,
        SetCellValue =              lib.SetCellValue,
        StartSelect =               lib.StartSelect,
        EndSelect =                 lib.EndSelect,
        SetVisibleSelection =       lib.SetVisibleSelection,
        UpdateSelect =              lib.UpdateSelect,
        SetInteractive =            lib.SetInteractive,
        SetColorRange =             lib.SetColorRange,
        SetRow =                    lib.SetRow,
        ClearCells =                lib.ClearCells,
        SortByColumn =              lib.SortByColumn,
        SetCallback =               lib.SetSheetCallback,
        UpdateFrameSizes =          lib.UpdateFrameSizes,
        ShowEditBox =               lib.ShowEditBox,
        SaveEditBox =               lib.SaveEditBox
	}
	table.insert(obj.sheets, sheet);
	return obj:Select(sheet.title);
end

function lib:DeleteSheet(title)
	local obj = self;
    local i = nil;
	if ( type(title) ~= "string" ) then error(("%s: Error in deleting sheet; Invalid sheet name - expected string, got %s!"):format(SPREADSHEET, type(title))); end
	for k, v in pairs(obj.sheets) do
		if ( v.title == title ) then
            if ( obj.selectedSheet == v ) then
                obj.selectedSheet = nil;
            end
			i = k;
			break;
		end		
	end
    if ( i~= nil ) then
        if ( obj.sheets[i].frame ~= nil ) then
            obj.sheets[i].frame:Hide();
            obj.sheets[i].frame:SetParent(nil); -- make sure the garbage collector frees up this frame and its children
            obj.sheets[i].header:Hide();
            obj.sheets[i].header:SetParent(nil);
        end
        table.remove(obj.sheets, i);
    end
end

function lib:Select(sheet)
	local obj = self;
	for k, v in pairs(obj.sheets) do
		if ( v.title == sheet ) then
            obj.selectedSheet = v;
            if ( obj.hasBookUI == true ) then
                obj.verticalScrollbar:SetValue(0);
            end
			return v;
		end
	end
	return nil;
end


--[[
LibSpreadsheet-1.0 HEADER FUNCTIONS
]]--

function lib:AddColumn(...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
	for i = 1, select("#", ...) do
		local column = select(i, ...);
        if ( type(column) == "table" or type(column) == "string" ) then
            local col = {
                name =              "",
                dataChanged =       true,
                width =             1,
                height =            1,
                color =             {1,1,1,1},
                background =        {0,0,0,0},
            };
            table.insert(obj.columns, col);
            local n = table.maxn(obj.columns);
            if ( obj:SetColumn(n, column) == false ) then
                table.remove(n);
            else
                obj.lastColumn = obj.lastColumn + 1;
            end
        else
            error(("%s: Error while adding column, expected string or table, got %s!"):format(SPREADSHEET, column));
        end
	end
end

function lib:SetColumn(column, properties)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = column
    if ( type(column) == "string" ) then
        c = tonumber(column, 36) - 9;
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while setting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    if ( sheet.columns[c].type == nil or sheet.columns[c].justify == nil ) then
        sheet.columns[c].type = "string";
        sheet.columns[c].justify = "LEFT";
    end
    if ( type(properties) == "string" ) then
        sheet.columns[c].name=properties;
        sheet.columns[c].dataChanged = true;
        return true;
    elseif ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            sheet.columns[c][key] = val;
            if ( key == "type" and properties.justify == nil ) then
                if ( val == "number" ) then
                    sheet.columns[c].justify = "RIGHT";
                else
                    sheet.columns[c].justify = "LEFT";
                end
            end
        end
        sheet.columns[c].dataChanged = true;
        return true;
    else
        error(("%s: Error while setting column properties, expected string or table, got %s!"):format(SPREADSHEET, type(properties)));
        return false;
    end
end


function lib:DeleteColumn(column)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while deleting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local c = nil;
    if ( type(column) == "string" ) then
        c = tonumber(column, 36) - 9;
        if ( c > 26 ) then c = 26; end
    else
        c = column;
    end
    if ( sheet.columns[c] == nil ) then error(("%s: Error while deleting column %s: No such column!"):format(SPREADSHEET, lib.cols[c])); end
    table.remove(sheet.columns, c);
    sheet.lastColumn = sheet.lastColumn - 1;
    for k, v in pairs(sheet.rows) do
        if ( v.cells[c] ~= nil ) then
            if ( v.cells[c].frame ~= nil ) then
                 v.cells[c].frame:Hide();
                 v.cells[c].frame:SetParent(nil);
             end
         end
        table.remove(v.cells, c);
    end
end



--[[
LibSpreadsheet-1.0 DATA FUNCTIONS
]]--

function lib:ClearCells()
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while clearing sheet, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    table.wipe(sheet.rows);
    if ( sheet.frame ~= nil ) then
        sheet.frame:Hide();
        sheet.frame:SetParent(nil);
        sheet.frame = nil;
    end
end


function lib:SortByColumn(...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local func = function() return false; end;
    local i = select("#", ...);
    local p = 0;
    local N = 0;
    while ( i > p ) do
        local col = select((i-p-1), ...);
        local dir = select((i-p), ...);
        local l = string.lower;
        if      ( dir == obj.parent.parent.sortAscending )         then dir = false;
        elseif  ( dir == obj.parent.parent.sortAscendingWithCase ) then dir = false; l = function(x) return x; end;
        elseif  ( dir == obj.parent.parent.sortDescending )        then dir = true;
        else                                                       dir = true;  l = function(x) return x; end;
        end
        p = p + 2;
        N = N + 1;
        if ( type(col) == "string" ) then
            col = tonumber(col, 36) - 9;
        end
        local n = N;
        local r = dir;
        local c = col;
        if ( sheet.columns[c].type == "number" ) then
            l = function(x) return x; end;
        end
        local myFunc = func;
        func = function(a,b) if ( a == nil or b == nil ) then return; end if (l(a.cells[c].data) or 0) > (l(b.cells[c].data) or 0) then return r; elseif (l(a.cells[c].data) or 0) == (l(b.cells[c].data) or 0) then return myFunc(a,b); end return not r; end;
    end
    sheet.sortDirection = math.fmod(sheet.sortDirection, 2) + 1;
    table.sort(sheet.rows, func);
    for key, row in pairs(sheet.rows) do
        row.positionChanged = true; -- Tell Render() that it needs to update row placements.
    end
end


function lib:AddRow(...)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while adding row, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local row = {};
    for i = 1, select("#", ...) do
		local data = select(i, ...);
        local cell = {};
        if ( type(data) ~= "table" ) then
            cell.data = data;
            cell.color = nil;
        else
            cell.data = data.data;
            cell.color = data.color or nil;
            cell.id = data.id or nil;
        end
        if ( sheet.parent.strict == true ) then
            local t =  sheet.columns[i].type;
            if ( t == "string" and type(cell.data) ~= "string" ) then error(("%s: Error while adding cell in column %s, expected string, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
            if ( t == "boolean" and type(cell.data) ~= "boolean" ) then error(("%s: Error while adding cell in column %s, expected boolean, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
            if ( ( t == "number" ) and type(cell.data)  ~= "number" ) then error(("%s: Error while adding cell in column %s, expected number, got %s!"):format(SPREADSHEET, lib.cols[i], type(cell.data))); end
        end
        if ( sheet.columns[i] ~= nil ) then
            if ( sheet.columns[i].type == "string" ) then
                lib.str:SetFontObject(lib.fonts.StringFont);
                if ( type(cell.data) == "boolean" ) then
                    if ( cell.data == true ) then cell.data = "true";
                    else cell.data = "false";
                    end
                end
                lib.str:SetFontObject(lib.fonts.StringFont);
                lib.str:SetText(cell.data);
                local w = lib.str:GetStringWidth();
                if ( w > sheet.columns[i].width ) then
                    sheet.columns[i].width= w;
                end
                
            elseif ( sheet.columns[i].type == "number" ) then
                local s = ("%%.%uf%%s"):format(sheet.columns[i].decimals or 0);
                lib.str:SetFontObject(lib.fonts.NumberFont);
                lib.str:SetText(string.format(s, cell.data, (sheet.columns[i].suffix or "")));
                local w = lib.str:GetStringWidth();
                if ( w > sheet.columns[i].width ) then
                    sheet.columns[i].width= w;
                end
            end
        end
        cell.dataChanged = true;
        table.insert(row, cell);
    end
    table.insert(sheet.rows, {background={0,0,0,0}, color={1,1,1,1}, cells = row});
    return table.maxn(sheet.rows);
end

function lib:SetRow(row, properties)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting row properties, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    row = tonumber(row);
    if ( sheet.rows[row] == nil ) then error(("%s: Error while setting row properties %s: No such row!"):format(SPREADSHEET, lib.cols[c])); end
    if ( type(properties) == "table" ) then
        for key, val in pairs(properties) do
            sheet.rows[row][key] = val;
        end
        for k, v in pairs(sheet.rows[row].cells) do
            v.dataChanged = true;
        end
        sheet.rows[row].positionChanged = true;
    end
end

function lib:DeleteRow(...)
	local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while deleting row, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local dels = {};
    for i = 1, select("#", ...) do
		local row = select(i, ...);
        if ( sheet.rows[row] ~= nil ) then
            table.insert(dels, row);
            if ( sheet.rows[row].frame ~= nil ) then
                sheet.rows[row].frame:Hide();
                sheet.rows[row].frame:SetParent(nil);
            end
        end
    end
    table.sort(dels, function(a,b) return (a > b); end);
    for k, v in pairs(dels) do
        table.remove(sheet.rows, v);
    end
    for k, row in pairs(sheet.rows) do
        row.positionChanged = true;
    end
end

function lib:SetCellValue(c, r, d)
    local t = (self.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while looking up column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = self;
    local column, row = sheet:ExtractLocation(c);
    if ( row ~= nil ) then
        d = r;
    else
        row = r;
    end
    if ( sheet.rows[row] ~= nil ) then
        if ( sheet.rows[row].cells[column] ~= nil ) then
            sheet.rows[row].cells[column].data = d;
            sheet.rows[row].cells[column].dataChanged = true;
        end
    end
end

function lib:ExtractLocation(str)
    local t = (self.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while looking up column, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = self;
    local col = nil;
    local row = nil;
    if ( type(str) == "string" ) then
        local c, r = (str):match("(%D+)(%d*)");
        if ( r ~= nil ) then
            row = r;
        end
        if ( string.len(c) == 1 ) then
            col = tonumber(c, 36) - 9;
        else
            for k, column in pairs(sheet.columns) do
                if ( column.name == c ) then
                    col = k;
                    break;
                end
            end
        end
    elseif ( type(str) == "number" ) then
        if ( str > 26 ) then str = math.fmod(str, 26); end
        if ( str < 1 ) then 
            while ( str < 1 ) do
                str = str + sheet.lastColumn;
            end
        end
        col = str;
    end
    return col, tonumber(row);
end



function lib:GoTo(column, row) -- takes arguments like: GoTo("B"), GoTo("B1"), GoTo("B", 3), GoTo(2,5), GoTo("Name", 3), GoTo("Level5");
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
	local c, r = sheet:ExtractLocation(column);
    if ( type(r) == "number" ) then
        row = r;
    end
    if ( sheet.columns[column] == nil ) then error(("%s: Error in GoTo(): No such column '%s'!"):format(SPREADSHEET, lib.cols[column or 27])); end
    if ( sheet.rows[row] == nil ) then error(("%s: Error in GoTo(): No such row '%u'!"):format(SPREADSHEET, row)); end
    sheet.currentColumn = column;
    sheet.currentRow = row;    
end

function lib:Delete(fromCol, fromRow, toCol, toRow)
	
end

function lib:Insert(...)
	
end

function lib:StrictMode(toggle)
	if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting strict mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.strict = toggle;
end



--[[
LibSpreadsheet-1.0 VISUAL HANDLERS
]]--


function lib:ShowSelector(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting selector mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showSelector = toggle;
end

function lib:ShowScrollbars(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting scrollbar visibility; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
	self.showScrollbars = toggle;
end

function lib:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in editing sheet: Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    sheet.isInteractive = toggle;
end

function lib:UpdateFrameSizes()
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in editing sheet: Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    
    for key, header in pairs (sheet.columns) do
        if ( header.dataChanged ) then
            lib.str:SetFontObject(lib.fonts.HeaderFont);
            lib.str:SetText((header.name or "??") .. ":");
            local height = math.ceil(lib.str:GetStringHeight());
            local width = math.ceil(lib.str:GetStringWidth());
            if ( header.height ~= height ) then header.height = height; end
            if ( header.width ~= width ) then header.width = width; end
            sheet.rowHeight = height;
            header.colwidth = width + (sheet.cellSpacing*2);
        end
    end
    local t = nil;
    for key, row in pairs (sheet.rows) do
        for key, cell in pairs (row.cells) do
            t = sheet.columns[key].type;
            if ( cell.dataChanged ) then
                if ( t == "number" ) then
                    lib.str:SetFontObject(lib.fonts.NumberFont);
                    local s = ("%%.%uf%%s"):format(sheet.columns[key].decimals or 0);
                    lib.str:SetText(string.format(s, cell.data or 0, (sheet.columns[key].suffix or "")));
                else
                    lib.str:SetFontObject(lib.fonts.StringFont);
                    lib.str:SetText(cell.data or "");
                end
                cell.height = math.ceil(lib.str:GetStringHeight());
                cell.width = math.ceil(lib.str:GetStringWidth() + (sheet.cellSpacing*2));
            end
            if ( cell.width > sheet.columns[key].colwidth ) then
                sheet.columns[key].colwidth = cell.width;
            end
            if ( cell.height > sheet.rowHeight ) then
                sheet.rowHeight = cell.height;
            end
        end
    end
end

function lib:BuildBookUI(obj)
    local t = (obj.type or "unknown");
    if ( t ~= "book" ) then error(("%s: Error in Book:Visualize(): Expected book, got %s!"):format(SPREADSHEET, t)); end
    
    obj.frame = CreateFrame("Frame", nil, UIParent);
    obj.outerScrollframe = CreateFrame("ScrollFrame", nil, obj.frame);
    obj.header = nil;
    obj.frame:SetPoint("CENTER", UIParent, "CENTER", 0,0);
    obj.frame:SetFrameStrata("DIALOG")
    obj.frame:EnableMouse(true)
    obj.innerFrame = nil;
    obj.frame:Hide();
    obj.offsetY = 0;
    obj.scrollFrame = CreateFrame("ScrollFrame", nil, obj.frame);
	obj.scrollFrame:SetPoint("TOPLEFT");
	obj.scrollFrame:SetPoint("BOTTOMRIGHT");
	obj.scrollFrame:EnableMouseWheel(true)
    obj.scrollFrame:EnableMouse(true) 
    obj.scrollFrame:SetAllPoints();
 
    obj.verticalScrollbar = CreateFrame("Slider", nil, obj.scrollFrame, "UIPanelScrollBarTemplate");
	obj.verticalScrollbar:SetPoint("TOPLEFT", obj.scrollFrame, "TOPRIGHT", 4, -16)
	obj.verticalScrollbar:SetPoint("BOTTOMLEFT", obj.scrollFrame, "BOTTOMRIGHT", 4, 16)
	obj.verticalScrollbar:SetMinMaxValues(0, 1000)
	obj.verticalScrollbar:SetValueStep(1)
	obj.verticalScrollbar:SetValue(0)
	obj.verticalScrollbar:SetWidth(16)
    obj.scrollFrame.scrollvalue = 0;
	obj.verticalScrollbar:Hide()
	obj.verticalScrollbar:SetScript("OnValueChanged", function(a,b) obj.scrollFrame:scrollFunc(b); end );
        
    obj.scrollFrame.onMouseWheel = function(self, value)
		local height, viewheight = obj.scrollFrame:GetHeight(), obj.innerFrame:GetHeight()
		
		if height > viewheight then
			obj.verticalScrollbar:Hide()
		else
            obj.verticalScrollbar:Show()
			local diff = height - viewheight
			local delta = 1
			if (value or 0) < 0 then
				delta = -1
			end
			obj.verticalScrollbar:SetValue(min(max(obj.scrollFrame.scrollvalue + delta*(obj.offsetY/(diff/45)),0), 1000))
		end
	end;
    
    
    obj.sheetSelector = CreateFrame("Frame", nil, obj.frame);
    obj.sheetSelector:SetFrameStrata("FULLSCREEN_DIALOG")
    obj.prevSheet = CreateFrame("Button", nil, obj.sheetSelector);
    obj.prevSheet:SetWidth(32);
    obj.prevSheet:SetHeight(32);
    obj.prevSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
    obj.prevSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
    obj.prevSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
    obj.prevSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    obj.prevSheet:SetScript("OnClick", function() obj:Leaf(-1); obj:Render(); end);
    obj.prevSheet:SetPoint("LEFT", obj.sheetSelector, "LEFT", 0, 0);
    
    
    obj.nextSheet = CreateFrame("Button", nil, obj.sheetSelector);
    obj.nextSheet:SetWidth(32);
    obj.nextSheet:SetHeight(32);
    obj.nextSheet:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up");
    obj.nextSheet:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down");
    obj.nextSheet:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled");
    obj.nextSheet:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight", "ADD");
    obj.nextSheet:SetScript("OnClick", function() obj:Leaf(1); obj:Render(); end);
    obj.nextSheet:SetPoint("RIGHT", obj.sheetSelector, "RIGHT", 0, 0);
    
    obj.sheetSelector:SetPoint("BOTTOMLEFT", obj.frame, "BOTTOMLEFT", 0, 0);
    obj.sheetSelector:SetPoint("TOPRIGHT", obj.frame, "BOTTOMRIGHT", 0, 32);
    
    obj.sheetSelectorStatus = CreateFrame("Frame", nil, obj.sheetSelector);
	obj.sheetSelectorStatus:SetBackdrop({
        bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
        edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
        tile = true, tileSize = 16, edgeSize = 16,
        insets = { left = 3, right = 3, top = 5, bottom = 3 }
    });
    obj.sheetSelectorStatus:SetAllPoints();
	obj.sheetSelectorStatus:SetBackdropColor(0.1, 0.1, 0.1, 0.5);
	obj.sheetSelectorStatus:SetBackdropBorderColor(0.4, 0.4, 0.4);
    
    obj.sheetSelectorText = obj.sheetSelectorStatus:CreateFontString(nil, "DIALOG", lib.fonts.HeaderFont);
    obj.sheetSelectorText:SetAllPoints(obj.sheetSelectorStatus);
    obj.sheetSelectorText:SetText("");
    obj.sheetSelectorText:SetTextColor(0.9,0.75,0.2);
    obj.sheetSelectorText:Show();
	obj.scrollFrame:SetScript("OnMouseWheel", obj.scrollFrame.onMouseWheel);
    obj.outerScrollframe:SetScript("OnShow", function(self) obj.scrollFrame:onMouseWheel(0); end);
    obj.scrollFrame.scrollFunc = function(self, value)
		local viewheight = obj.scrollFrame:GetHeight()
		local height = obj.innerFrame:GetHeight()
		local offset

		if viewheight > height then
			offset = 0
            obj.verticalScrollbar:Hide()
		else
            obj.verticalScrollbar:SetMinMaxValues(0, height-viewheight)
            obj.verticalScrollbar:Show()
			offset = value;
		end
		obj.innerFrame:ClearAllPoints()
		obj.innerFrame:SetPoint("TOPLEFT", 0, offset)
		obj.innerFrame:SetPoint("TOPRIGHT", 0, offset)
		
		obj.scrollFrame.scrollvalue = value
    end;
    
    obj.editbox = CreateFrame("EditBox", nil, obj.frame, "InputBoxTemplate");
    obj.editbox:SetFrameLevel(999);
	obj.editbox:SetAutoFocus(true);
	obj.editbox:SetFontObject(lib.fonts.StringFont);
	obj.editbox:SetScript("OnEscapePressed", function(me) me:Hide(); end);
    obj.editbox:SetScript("OnHide", function(me) if ( me.caller ~= nil ) then me.caller.label:Show();end end);
	obj.editbox:SetScript("OnEnterPressed",  function(me) if ( me.caller ~= nil ) then me.caller.sheet:SaveEditBox(me.caller, me:GetText()); end end);
	obj.editbox:SetTextInsets(0, 0, 3, 3);
	obj.editbox:SetMaxLetters(256);
    obj.editbox:Hide();
    obj.hasBookUI = true;
end

function lib:Hide()
    local obj = self;
    if ( obj.frame ~= nil ) then
        obj.frame:Hide();
    end
end


function lib:Render()
    local obj = self;
    if ( obj.hasBookUI == false ) then
        obj.parent:BuildBookUI(obj);
    end
    local sheet = obj.selectedSheet;
    sheet:UpdateFrameSizes();
    if ( obj.innerFrame ~= nil ) then
        obj.innerFrame:Hide();
        obj.header:Hide();
    end
    if ( sheet.frame == nil ) then
        sheet.frame = CreateFrame("Frame", nil, obj.scrollFrame);
        sheet.frame:SetAllPoints(obj.frame);
        sheet.frame.placeholder = CreateFrame("Frame", nil, sheet.frame);
        sheet.frame.placeholder:SetPoint("TOPLEFT", sheet.frame, "TOPLEFT", -1, 0);
        sheet.frame.placeholder:SetPoint("BOTTOMRIGHT", sheet.frame, "TOPRIGHT", -1, -2);
    end
    if ( sheet.header == nil ) then
        sheet.header = CreateFrame("Frame", nil, obj.frame);
        sheet.header:SetPoint("TOPLEFT", obj.frame, "TOPLEFT", 0, 0);
        sheet.header:SetPoint("BOTTOMRIGHT", obj.frame, "TOPLEFT", 0, -20);
    end
    
    obj.innerFrame = sheet.frame;
    obj.header = sheet.header;
    obj.scrollFrame:SetScrollChild(sheet.frame);
    obj.verticalScrollbar:Hide();
    sheet.frame:Show();
    sheet.header:Show();
    
    local offset = 0;
    for key, cell in pairs (sheet.columns) do
        if ( cell.frame == nil ) then
            local cellFrame = CreateFrame("button", nil, obj.header);
            cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG", lib.fonts.HeaderFont);
            cellFrame.label:SetFontObject(lib.fonts.HeaderFont);
            cellFrame.label:SetPoint("TOPLEFT", cellFrame, "TOPLEFT", (sheet.cellSpacing/2), -(sheet.cellSpacing/2));
            cellFrame.label:SetPoint("BOTTOMRIGHT", cellFrame, "BOTTOMRIGHT", -(sheet.cellSpacing/2), (sheet.cellSpacing/2));
            cell.frame = cellFrame;
        end
        local cellFrame = cell.frame;
        if ( cell.dataChanged ) then
            if ( cell.sort == true ) then
                local k = key;
                cellFrame:SetScript("OnClick", 
                    function() 
                        local l = true;
                        if ( sheet.callbacks["OnUserSortRequested"] ~= nil ) then
                            if ( type(sheet.callbacks["OnUserSortRequested"]) == "function" ) then
                                l = (sheet.callbacks["OnUserSortRequested"])(lib.cols[k]);
                            end
                        end
                        if ( l ) then
                            sheet:SortByColumn(k, sheet.sortDirection);
                            obj:Render();
                            if ( sheet.callbacks["OnUserSortProcessed"] ~= nil ) then
                                if ( type(sheet.callbacks["OnUserSortProcessed"]) == "function" ) then
                                    (sheet.callbacks["OnUserSortProcessed"])(lib.cols[k]);
                                end
                            end
                        end
                    end
                );
            end
            cellFrame.label:SetText((cell.name or "??")..":");
            cellFrame.label:SetTextColor(0.9,0.75,0.2);
            cellFrame.label:SetJustifyH(cell.justify);
        end
        cellFrame:SetWidth(cell.colwidth);
        cellFrame:SetHeight(sheet.rowHeight);
        cellFrame:SetPoint("TOPLEFT", obj.header, "TOPLEFT", offset, 0);
        offset = offset + cell.colwidth + 2;
    end    
    
    local offsetY = sheet.cellSpacing/2;
    local prevRow = sheet.frame.placeholder;
    for rkey, row in pairs (sheet.rows) do
        local offsetX = 0;
        if ( row.frame == nil ) then
            row.frame = CreateFrame("Frame", "Row" .. rkey, sheet.frame);
            row.frame:ClearAllPoints()
            row.frame:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0,  -sheet.cellSpacing);
            row.frame:SetPoint("TOPRIGHT", prevRow, "BOTTOMRIGHT", 0, -sheet.cellSpacing);
            row.frame:SetFrameLevel(rkey*10);
            row.frame:SetHeight(sheet.rowHeight + sheet.cellSpacing);
            row.frame.texture = row.frame:CreateTexture();
            row.frame.texture:SetTexture(0,0,0,0);
            row.frame.texture:SetAllPoints(row.frame)
            local myPrevRow = prevRow;
        end
        if ( row.positionChanged ) then
            row.frame:ClearAllPoints()
            row.frame:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT", 0,   -sheet.cellSpacing);
            row.frame:SetPoint("TOPRIGHT", prevRow, "BOTTOMRIGHT", 0,  -sheet.cellSpacing);
            row.frame:SetHeight(sheet.rowHeight + sheet.cellSpacing);
            row.frame.texture:SetTexture(row.background[1], row.background[2], row.background[3], row.background[4]);
            local myPrevRow = prevRow;
            row.positionChanged = false;
        end
        prevRow = row.frame;
        local i = 0;
        for key, cell in pairs(row.cells) do
            i = i + 1;
            if ( cell.frame == nil ) then
                local cellFrame = CreateFrame("Button", nil, row.frame);
                cellFrame.label = cellFrame:CreateFontString(nil, "DIALOG", lib.fonts.StringFont);
                cellFrame.label:SetPoint("TOPLEFT", cellFrame, "TOPLEFT", (sheet.cellSpacing), -(sheet.cellSpacing/2));
                cellFrame.label:SetPoint("BOTTOMRIGHT", cellFrame, "BOTTOMRIGHT", -(sheet.cellSpacing), (sheet.cellSpacing/2));
                cellFrame.tex = cellFrame:CreateTexture(nil, "OVERLAY");
                cellFrame.tex:SetAllPoints();
                cellFrame.tex:SetBlendMode("ADD");
                cellFrame.back = cellFrame:CreateTexture(nil, "BACKGROUND");
                cellFrame.back:SetAllPoints();
                cell.frame = cellFrame;
            end
            local cellFrame = cell.frame;
            cellFrame:SetPoint("TOPLEFT", row.frame, "TOPLEFT", offsetX, -(sheet.cellSpacing/2));
            cellFrame:SetWidth(sheet.columns[i].colwidth);
            cellFrame:SetHeight(sheet.rowHeight);
            cellFrame.column = i;
            cellFrame.row = rkey;
            cellFrame.sheet = sheet;
            cellFrame.cell = cell;
            offsetX = offsetX + sheet.columns[i].colwidth + sheet.cellSpacing;
            
            if ( cell.dataChanged ) then
                cellFrame.label:SetJustifyH(sheet.columns[i].justify);
                cellFrame.label:SetTextColor(sheet.columns[i].color[1], sheet.columns[i].color[2], sheet.columns[i].color[3]);
                cellFrame.back:SetTexture(sheet.columns[i].background[1], sheet.columns[i].background[2], sheet.columns[i].background[3], sheet.columns[i].background[4]);
                if ( cell.color ~= nil ) then
                    cellFrame.label:SetTextColor(cell.color[1] or cell.color.r, cell.color[2] or cell.color.g, cell.color[3] or cell.color.b);
                end
                if ( cell.background ~= nil ) then
                    cellFrame.back:SetTexture(cell.background[1],cell.background[2], cell.background[3], cell.background[4]);
                end
                if ( sheet.isInteractive or sheet.columns[i].interactive ) then
                    cellFrame:SetScript("OnMouseDown", function(caller, event) sheet:StartSelect(cellFrame, event); end);
                    cellFrame:SetScript("OnMouseUp", function(caller, event) sheet:EndSelect(cellFrame, event); end);
                    cellFrame:SetScript("OnEnter", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                    cellFrame:SetScript("OnLeave", function(caller, event) sheet:UpdateSelect(cellFrame); end);
                    cellFrame:SetScript("OnDoubleclick", function(caller, event) sheet:ShowEditBox(cellFrame); end);
                end
                if ( sheet.columns[i].onClick ~= nil ) then
                    local func = sheet.columns[i].onClick;
                    cellFrame:SetScript("OnClick", function(caller, event) func(cell.id, cell.data, cellFrame); end);

                end
                
                if ( sheet.columns[i].type == "number" ) then
                    cellFrame.label:SetFontObject(lib.fonts.NumberFont);
                    if ( sheet.columns[i].decimals ~= nil ) then
                        local s = ("%%.%uf%%s"):format(sheet.columns[i].decimals);
                        cellFrame.label:SetText(string.format(s, cell.data or 0, (sheet.columns[i].suffix or "")));
                    else
                        cellFrame.label:SetText(string.format("%i%s", cell.data or 0, (sheet.columns[i].suffix or "")));
                    end
                    if ( sheet.columns[i].colorRange ~= nil ) then
                        if ( sheet.columns[i].minval ~= nil and sheet.columns[i].maxval ~= nil ) then
                            local colors = table.maxn(sheet.columns[i].colorRange);
                            local where = ( ( (cell.data - sheet.columns[i].minval)  / (sheet.columns[i].maxval - sheet.columns[i].minval)) * (colors-1) ) + 1;
                            local p, n = math.floor(where), math.ceil(where);
                            local r = (where - p) * 100;
                            if (  sheet.columns[i].colorRange[p] == nil ) then
                                where = ((colors-1)/2) + 1
                                p, n = math.floor(where), math.ceil(where);
                                r = 0;
                            end
                            local color = { 
                                (  ( sheet.columns[i].colorRange[p][1] * (100-r) ) + ( sheet.columns[i].colorRange[n][1] * (r) ) ) / 100,
                                (  ( sheet.columns[i].colorRange[p][2] * (100-r) ) + ( sheet.columns[i].colorRange[n][2] * (r) ) ) / 100,
                                (  ( sheet.columns[i].colorRange[p][3] * (100-r) ) + ( sheet.columns[i].colorRange[n][3] * (r) ) ) / 100
                                };
                            cellFrame.label:SetTextColor(color[1], color[2], color[3]);
                            if ( sheet.columns[i].dominant ) then
                                for k, v in pairs(sheet.rows[rkey].cells) do
                                    v.color = {color[1], color[2], color[3]};
                                    if ( v.frame ) then
                                        v.frame.label:SetTextColor(color[1], color[2], color[3]);
                                    end
                                end
                            end
                        end
                    end
                elseif ( sheet.columns[i].type == "string" ) then
                    cellFrame.label:SetFontObject(lib.fonts.StringFont);
                    cellFrame.label:SetText(cell.data or "");
                elseif ( sheet.columns[i].type == "boolean" ) then
                    cellFrame.label:SetFontObject(lib.fonts.StringFont);
                    if ( cell.data == true ) then cellFrame.label:SetText("true");
                    else cellFrame.label:SetText("false"); end
                end
            end
        end
        offsetY = offsetY + sheet.rowHeight + 2 + sheet.cellSpacing;
    end
    obj.offsetY = offsetY;
    obj.scrollFrame:SetPoint("TOPLEFT", obj.header, "BOTTOMLEFT", 0, 0);
    obj.scrollFrame:SetPoint("BOTTOMRIGHT", obj.frame, "BOTTOMRIGHT", -16, 16);
    obj.verticalScrollbar:SetMinMaxValues(0, offsetY);
    obj.verticalScrollbar:SetValue(0);
    obj.innerFrame:SetHeight(offsetY);
    obj.scrollFrame:scrollFunc(0);
    obj.scrollFrame:onMouseWheel(0);
    obj.sheetSelector:Hide();
    if ( obj.showSelector == true ) then
        obj:Leaf(0);
        obj.scrollFrame:SetPoint("BOTTOMRIGHT", obj.frame, "BOTTOMRIGHT", -16, 48);
        obj.sheetSelector:Show();
    end
    return obj.frame;
end



function lib:SetInteractive(toggle)
    if ( type(toggle) ~= "boolean" ) then error(("%s: Error in setting interactive mode; Expected boolean, got %s!"):format(SPREADSHEET, type(toggle))); end
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in editing sheet: Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    sheet.isInteractive = toggle;
end

function lib:SetColorRange(column, ...)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error while setting color range, expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    if ( type(column) == "string" ) then
        column = tonumber(column, 36) - 9;
    end
    if ( sheet.columns[column] == nil ) then error(("%s: Error while setting color range for column %s: No such column!"):format(SPREADSHEET, lib.cols[column])); end
    sheet.columns[column].colorRange = {};
    for i = 1, select("#", ...) do
		local color = select(i, ...);
        table.insert(sheet.columns[column].colorRange, color);
    end
end


function lib:Leaf(direction)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "book" ) then error(("%s: Error in GoTo(): Expected book, got %s!"):format(SPREADSHEET, t)); end
    local c = 1;
    for k, v in pairs(obj.sheets) do
        if ( v.title == obj.selectedSheet.title ) then
            c = k;
            break;
        end
    end
    if ( direction == 1 ) then
        c = c + 1;
        if ( obj.sheets[c] ~= nil ) then
            obj.selectedSheet = obj.sheets[c];
        end
    elseif ( direction == -1 ) then
        c = c - 1;
        if ( obj.sheets[c] ~= nil ) then
            obj.selectedSheet = obj.sheets[c];
        end
    end
    if ( c < 2 ) then
        obj.prevSheet:Disable();
    else
        obj.prevSheet:Enable();
    end
    if ( c == table.maxn(obj.sheets) ) then
        obj.nextSheet:Disable();
    else
        obj.nextSheet:Enable();
    end
    obj.sheetSelectorText:SetText(obj.title .. ": " .. obj.selectedSheet.title);
    obj.verticalScrollbar:SetValue(0);
end

function lib:SetVisibleSelection(fromCol, fromRow, toCol, toRow)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in GoTo(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    if ( type(fromCol) == "number" ) then
        sheet.selectStartColumn = fromCol;
        sheet.selectStartRow = fromRow;
    elseif ( type(fromCol) == "string" ) then
        local c, r = sheet:ExtractLocation(fromCol);
        if ( type(r) == "number" ) then
            sheet.selectStartColumn = c;
            sheet.selectStartRow = r;
            toRow = toCol;
            toCol = fromRow;
        else
            sheet.selectStartColumn = c;
            sheet.selectStartRow = fromRow;
        end
    end
    if ( type(toCol) == "number" ) then
        sheet.selectEndColumn = toCol;
        sheet.selectEndRow = toRow;
    elseif ( type(toCol) == "string" ) then
        local c, r = sheet:ExtractLocation(toCol);
        if ( type(r) == "number" ) then
            sheet.selectEndColumn = c;
            sheet.selectEndRow = r;
        else
            sheet.selectEndColumn = c;
            sheet.selectEndRow = toRow;
        end
    end
    sheet.selecting = true;
    sheet:UpdateSelect(nil);
    sheet.selecting = false;
end


function lib:StartSelect(caller, event)
    local sheet = self;
    sheet.selecting = true;
    sheet.selectStartColumn = caller.column;
    sheet.selectStartRow = caller.row;
    if ( event == "RightButton" ) then
        sheet.selectStartColumn = nil;
        sheet.selectStartRow = nil;
        sheet.selectEndColumn = nil;
        sheet.selectEndRow = nil;
    end
    sheet:UpdateSelect(caller);
end

function lib:EndSelect(caller, event)
    local sheet = self;
    sheet.selecting = false;
    if ( event == "LeftButton" ) then
        if not ( sheet.selectStartColumn == sheet.selectEndColumn and sheet.selectStartRow == sheet.selectEndRow ) then
            if ( sheet.callbacks['OnSelectionChange'] ~= nil ) then
                if ( type(sheet.callbacks['OnSelectionChange']) == "function" ) then
                    (sheet.callbacks['OnSelectionChange'])(math.min(sheet.selectStartColumn, sheet.selectEndColumn), math.min(sheet.selectStartRow, sheet.selectEndRow), math.max(sheet.selectStartColumn, sheet.selectEndColumn), math.max(sheet.selectStartRow, sheet.selectEndRow));
                end
            end
        end
    end
end

function lib:UpdateSelect(caller)
    local sheet = self;
    if ( sheet.selecting or false == true ) then
        if ( caller ~= nil ) then
            sheet.selectEndColumn = caller.column;
            sheet.selectEndRow = caller.row;
        end
        for x = 1, table.maxn(sheet.rows) do
            for y = 1, table.maxn(sheet.columns) do
                sheet.rows[x].cells[y].frame.tex:SetTexture(0,0,0,0);
            end
        end
        if ( sheet.selectEndRow == nil ) then sheet.selectEndRow  = sheet.selectStartRow ; end
        if ( sheet.selectStartRow ~= nil ) then
            for x = math.min(sheet.selectStartRow, sheet.selectEndRow), math.max(sheet.selectStartRow, sheet.selectEndRow) do
                for y = math.min(sheet.selectStartColumn, sheet.selectEndColumn), math.max(sheet.selectStartColumn, sheet.selectEndColumn) do
                    sheet.rows[x].cells[y].frame.tex:SetTexture(0.3,0.2,1,0.25);
                end
            end
        end
    end
end

function lib:SetSheetCallback(event, func)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in SetCallback(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local found = false;
    for k,v in pairs(lib.callbackEvents) do
        if ( v == event ) then
            found = true;
            break;
        end
    end
    if ( found == true ) then
        if ( type(func) == "function" ) then
            sheet.callbacks[event] = func;
        else
            error(("%s: Error in SetCallback(): expected function, got %s!"):format(SPREADSHEET, type(func)));
        end
    else
        error(("%s: Error in SetCallback(): No such event, '%s'!"):format(SPREADSHEET, event));
    end
end

function lib:ShowEditBox(cellFrame)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in ShowEditBox(): Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local ret = true;
    if ( sheet.parent.editbox:IsVisible() ) then
        sheet.parent.editbox:Hide();
    end
    if ( sheet.callbacks['OnUserInputRequested'] ~= nil ) then
        if ( type(sheet.callbacks['OnUserInputRequested']) == "function" ) then
            ret = (sheet.callbacks['OnUserInputRequested'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data);
        end
    end
    if ( ret ) then
        cellFrame.label:Hide();
        sheet.parent.editbox.caller = cellFrame;
        sheet.parent.editbox:SetParent(cellFrame);
        sheet.parent.editbox:SetAllPoints();
        sheet.parent.editbox:SetHeight(15);
        sheet.parent.editbox:Show();
    end
end

function lib:SaveEditBox(cellFrame, text)
    local obj = self;
    local t = (obj.type or "unknown");
    if ( t ~= "sheet" ) then error(("%s: Error in SaveEditBox: Expected sheet, got %s!"):format(SPREADSHEET, t)); end
    local sheet = obj;
    local ret = true;
    if ( sheet.callbacks['OnUserInputProcessed'] ~= nil ) then
        if ( type(sheet.callbacks['OnUserInputProcessed']) == "function" ) then
            ret = (sheet.callbacks['OnUserInputProcessed'])(cellFrame.col, cellFrame.row, cellFrame.cell.id, cellFrame.cell.data, text);
        end
    end
    if ( ret ) then
        cellFrame.cell.data = text;
        cellFrame.cell.dataChanged = true;
        obj.parent:Render();
    end
    sheet.parent.editbox:ClearAllPoints();
    sheet.parent.editbox:SetText("");
end


--[[
LibSpreadsheet-1.0 EXAMPLES
]]--

function LibSpreadTest()
    local lss = LibStub("LibSpreadsheet-1.0", true);                    -- Load the lib
    local book = lss:Book("LibSpreadsheet examples");                   -- Make a new book
    local sheet = book:Add("My stuff");                                 -- Add a sheet to the book
    book:StrictMode(true);
    sheet.isInteractive = true;
    book:ShowSelector(true);
    sheet:AddColumn(
        {name="Name", sort=true, background={0.7,0.9,0.8,0.3}},
        "Class",
        {name="Level", type="number", minval=1, maxval=80, sort=true, background={1,0.9,0.2,0.3}},
        {name="Score", type="number", decimals=2, suffix="%", sort=true},
        {name="Gnomish", type="boolean", color={0,1,1}}
        );  -- Set some headers
    sheet:AddRow("Apricot", "Warrior", 79, 95.5, true);                     -- Add some data
    sheet:AddRow("Bumblebee", "|cffFF4499Priest|r", 27, 82.5, false);                      -- Add another row
    sheet:AddRow("Charlie", "Gnome?", 50, 23, true);                     -- Add some data
    sheet:AddRow("Doris", "Warriorwarriorwarrior", 60, 31.34, true);                     -- Add some data
    sheet:AddRow("Evelyn", "|cffFF4499Priest|r", 80, 45.7, false);                      -- Add another row
    sheet:AddRow("Frank", "Gnome....", 1, 67.5, true);                     -- Add some data
    sheet:AddRow("George", "Warrior", 40, 8.4, true);                     -- Add some data
    sheet:AddRow("Henry", "|cffFF4499Priest|r", 75, 2.5, false);                      -- Add another row
    sheet:AddRow("Ian", "Gnome....", 12, 11, true);                     -- Add some data
    sheet:SetColorRange("C", {1,0,0}, {1,1,0}, {0,1,0});
    --sheet:SortByColumn("D", lss.sortDescending, "A", lss.sortAscending);
    sheet = book:Add("Other stuff");
    sheet:AddColumn("Eeny", "Meeny", "Miny", "Mo");
    sheet:AddRow("Meeny", "Miny", "Mo", "Eeny");
    sheet:AddRow("Miny", "Mo", "Eeny", "Meeny");
    sheet:AddRow("Mo", "Eeny", "Meeny", "Miny");
    
    sheet = book:Select("My stuff");
    sheet:SetCellValue("B", 2, "Mooo");
    sheet:SetCellValue(3, 3, 77);
    --sheet:DeleteColumn("B");
    local frame = book:Render();                                        -- Create the UI
    local window = CreateFrame("Frame", nil, UIParent, "DialogBoxFrame");
    window:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
    window:SetWidth(500);
    window:SetHeight(300);
    window:RegisterForDrag("LeftButton");
    window:SetScript("OnDragStart",function()  window:StartMoving(); end);
    window:SetScript("OnDragStop", function()  window:StopMovingOrSizing(); end);
    window:SetMovable(true);
    window:EnableMouse(true);
    frame:SetParent(window);
    frame:SetPoint("TOPLEFT", window, "TOPLEFT", 20, -20);
    frame:SetPoint("BOTTOMRIGHT", window, "BOTTOMRIGHT", -10, 50);
    frame:Show();
    window:Show();
end

